VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsToolTip"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
  ' ==================================================================================================
  ' clsToolTip.cls
  '
  ' TimoSoft ToolTip class
  ' written by Timo "TimoSoft" Kunze (Timo@TimoSoft-Software.com)
  '
  ' This is an implementation of the Windows ToolTip control including Unicode support.
  '
  ' release history:
  ' ----------------
  '  1.3 (08/20/2004) added hBitmap support
  '                   bugfixes
  '  1.2 (03/21/2004) removed unused declarations
  '  1.1 (02/03/2004) added Attach() and Detach() methods
  '                   Property Get Unicode() now uses IsWindowUnicode()
  '                   some small code cleanups
  '  1.0 (01/29/2004) initial release
  '
  ' known issues:
  ' -------------
  '  - hBitmap support doesn't work with BalloonStyle = True
  '  - hBitmap support doesn't draw the close button if ShowCloseButton = True
  '  - NeedText() supports ANSI text only, because Windows doesn't seem to make use of TTN_NEEDTEXTW.
  '         -> probably because the parent window is ANSI
  '  - NeedText(): StoreData doesn't have any effect.
  '
  ' --------------------------------------------------------------------------------------------------
  ' visit http://www.TimoSoft-Software.de
  ' ==================================================================================================

  #Const CustomDraw = True
  #Const Debuging = False
  #Const FireEvents = True
  #Const FullFeatured = True


  Private Enum CommonControlsConsts
    #If Debuging Then
      ICC_ANIMATE_CLASS = &H80
      ICC_COOL_CLASSES = &H400
      ICC_DATE_CLASSES = &H100
      ICC_HOTKEY_CLASS = &H40
      ICC_INTERNET_CLASSES = &H800
      ICC_LINK_CLASS = &H8000
      ICC_LISTVIEW_CLASSES = &H1
      ICC_NATIVEFNTCTL_CLASS = &H2000
      ICC_PAGESCROLLER_CLASS = &H1000
      ICC_PROGRESS_CLASS = &H20
      ICC_STANDARD_CLASSES = &H4000
      ICC_TAB_CLASSES = &H8
      ICC_TREEVIEW_CLASSES = &H2
      ICC_UPDOWN_CLASS = &H10
      ICC_USEREX_CLASSES = &H200
      ICC_WIN95_CLASSES = &HFF
    #End If
    ICC_BAR_CLASSES = &H4
  End Enum

  Private Enum ExWindowStyleConsts
    WS_EX_ACCEPTFILES = &H10
    WS_EX_APPWINDOW = &H40000
    WS_EX_CLIENTEDGE = &H200
    WS_EX_COMPOSITED = &H2000000
    WS_EX_CONTEXTHELP = &H400
    WS_EX_CONTROLPARENT = &H10000
    WS_EX_DLGMODALFRAME = &H1
    WS_EX_LAYERED = &H80000
    WS_EX_LAYOUTRTL = &H400000
    WS_EX_LEFT = &H0
    WS_EX_LEFTSCROLLBAR = &H4000
    WS_EX_LTRREADING = &H0
    WS_EX_MDICHILD = &H40
    WS_EX_NOACTIVATE = &H8000000
    WS_EX_NOINHERITLAYOUT = &H100000
    WS_EX_NOPARENTNOTIFY = &H4
    WS_EX_RIGHT = &H1000
    WS_EX_RIGHTSCROLLBAR = &H0
    WS_EX_RTLREADING = &H2000
    WS_EX_STATICEDGE = &H20000
    WS_EX_TOOLWINDOW = &H80
    WS_EX_TOPMOST = &H8
    WS_EX_TRANSPARENT = &H20
    WS_EX_WINDOWEDGE = &H100
    WS_EX_OVERLAPPEDWINDOW = (WS_EX_WINDOWEDGE Or WS_EX_CLIENTEDGE)
    WS_EX_PALETTEWINDOW = (WS_EX_WINDOWEDGE Or WS_EX_TOOLWINDOW Or WS_EX_TOPMOST)
  End Enum

  Private Enum ToolTipStyleConsts
    TTS_ALWAYSTIP = &H1
    TTS_BALLOON = &H40
    TTS_CLOSE = &H80
    TTS_NOANIMATE = &H10
    TTS_NOFADE = &H20
    TTS_NOPREFIX = &H2
  End Enum

  Private Enum WindowStyleConsts
    WS_BORDER = &H800000
    WS_CAPTION = &HC00000
    WS_CHILD = &H40000000
    WS_CLIPCHILDREN = &H2000000
    WS_CLIPSIBLINGS = &H4000000
    WS_DISABLED = &H8000000
    WS_DLGFRAME = &H400000
    WS_GROUP = &H20000
    WS_HSCROLL = &H100000
    WS_MAXIMIZE = &H1000000
    WS_MAXIMIZEBOX = &H10000
    WS_MINIMIZE = &H20000000
    WS_MINIMIZEBOX = &H20000
    WS_OVERLAPPED = &H0
    WS_POPUP = &H80000000
    WS_SYSMENU = &H80000
    WS_TABSTOP = &H10000
    WS_THICKFRAME = &H40000
    WS_VISIBLE = &H10000000
    WS_VSCROLL = &H200000
    WS_CHILDWINDOW = WS_CHILD
    WS_ICONIC = WS_MINIMIZE
    WS_OVERLAPPEDWINDOW = (WS_OVERLAPPED Or WS_CAPTION Or WS_SYSMENU Or WS_THICKFRAME Or WS_MINIMIZEBOX Or WS_MAXIMIZEBOX)
    WS_POPUPWINDOW = (WS_POPUP Or WS_BORDER Or WS_SYSMENU)
    WS_SIZEBOX = WS_THICKFRAME
    WS_TILED = WS_OVERLAPPED
    WS_TILEDWINDOW = WS_OVERLAPPEDWINDOW
  End Enum


  #If CustomDraw Or Debuging Then
    Public Enum BitmapPositionConstants
      bpLeft
      bpTop
      bpRight
      bpBottom
    End Enum
  #End If

  #If FullFeatured Or Debuging Then
    Public Enum ToolTipTitleIconConstants
      tttiNone = 0
      tttiInfo = 1
      tttiWarning = 2
      tttiError = 3
    End Enum
  #End If


  Private Const CLR_INVALID = -1
  Private Const LPSTR_TEXTCALLBACKA = -1
  Private Const LPSTR_TEXTCALLBACKW = -1
  Private Const NM_FIRST = 0
  Private Const TTN_FIRST = -520
  #If Debuging Then
    Private Const TTN_LAST = -549
  #End If
  Private Const WM_USER = &H400
  ' classname for buttons
  Private Const WC_TOOLTIP = "tooltips_class32"

  ' common controls messages
  #If FullFeatured Or Debuging Then
    Private Const CCM_FIRST = &H2000
    Private Const CCM_DPISCALE = (CCM_FIRST + &HC)
    Private Const CCM_GETCOLORSCHEME = (CCM_FIRST + 3)
    Private Const CCM_GETDROPTARGET = (CCM_FIRST + 4)
    Private Const CCM_GETUNICODEFORMAT = (CCM_FIRST + 6)
    Private Const CCM_GETVERSION = (CCM_FIRST + &H8)
    Private Const CCM_SETBKCOLOR = (CCM_FIRST + 1)
    Private Const CCM_SETCOLORSCHEME = (CCM_FIRST + 2)
    Private Const CCM_SETNOTIFYWINDOW = (CCM_FIRST + &H9)
    Private Const CCM_SETUNICODEFORMAT = (CCM_FIRST + 5)
    Private Const CCM_SETVERSION = (CCM_FIRST + &H7)
    Private Const CCM_SETWINDOWTHEME = (CCM_FIRST + &HB)
  #End If

  ' constants used with CustomDraw
  #If CustomDraw Or Debuging Then
    Private Const CDDS_MAPPART = &H5
    Private Const CDDS_POSTERASE = &H4
    Private Const CDDS_PREERASE = &H3
    Private Const CDDS_ITEM = &H10000
    Private Const CDDS_PREPAINT = &H1
    Private Const CDDS_POSTPAINT = &H2
    Private Const CDDS_ITEMPOSTERASE = (CDDS_ITEM Or CDDS_POSTERASE)
    Private Const CDDS_ITEMPOSTPAINT = (CDDS_ITEM Or CDDS_POSTPAINT)
    Private Const CDDS_ITEMPREERASE = (CDDS_ITEM Or CDDS_PREERASE)
    Private Const CDDS_ITEMPREPAINT = (CDDS_ITEM Or CDDS_PREPAINT)
    Private Const CDDS_SUBITEM = &H20000
    Private Const CDIS_CHECKED = &H8
    Private Const CDIS_DEFAULT = &H20
    Private Const CDIS_DISABLED = &H4
    Private Const CDIS_FOCUS = &H10
    Private Const CDIS_GRAYED = &H2
    Private Const CDIS_HOT = &H40
    Private Const CDIS_INDETERMINATE = &H100
    Private Const CDIS_MARKED = &H80
    Private Const CDIS_SELECTED = &H1
    Private Const CDIS_SHOWKEYBOARDCUES = &H200
    Private Const CDRF_DODEFAULT = &H0
    Private Const CDRF_NEWFONT = &H2
    Private Const CDRF_NOTIFYITEMDRAW = &H20
    Private Const CDRF_NOTIFYPOSTERASE = &H40
    Private Const CDRF_NOTIFYPOSTPAINT = &H10
    Private Const CDRF_NOTIFYSUBITEMDRAW = &H20
    Private Const CDRF_SKIPDEFAULT = &H4
  #End If

  ' tooltip messages
  #If Debuging Then
    Private Const TTM_ADJUSTRECT = (WM_USER + 31)
    Private Const TTM_ENUMTOOLSA = (WM_USER + 14)
    Private Const TTM_ENUMTOOLSW = (WM_USER + 58)
    Private Const TTM_GETBUBBLESIZE = (WM_USER + 30)
    Private Const TTM_GETCURRENTTOOLA = (WM_USER + 15)
    Private Const TTM_GETCURRENTTOOLW = (WM_USER + 59)
    Private Const TTM_GETTOOLCOUNT = (WM_USER + 13)
    Private Const TTM_GETTOOLINFOA = (WM_USER + 8)
    Private Const TTM_GETTOOLINFOW = (WM_USER + 53)
    Private Const TTM_HITTESTA = (WM_USER + 10)
    Private Const TTM_HITTESTW = (WM_USER + 55)
    Private Const TTM_NEWTOOLRECTA = (WM_USER + 6)
    Private Const TTM_NEWTOOLRECTW = (WM_USER + 52)
    Private Const TTM_POPUP = (WM_USER + 34)
    Private Const TTM_RELAYEVENT = (WM_USER + 7)
    Private Const TTM_SETTOOLINFOA = (WM_USER + 9)
    Private Const TTM_SETTOOLINFOW = (WM_USER + 54)
    Private Const TTM_SETWINDOWTHEME = CCM_SETWINDOWTHEME
    Private Const TTM_UPDATE = (WM_USER + 29)
    Private Const TTM_WINDOWFROMPOINT = (WM_USER + 16)
  #End If
  Private Const TTM_ACTIVATE = (WM_USER + 1)
  Private Const TTM_ADDTOOLA = (WM_USER + 4)
  Private Const TTM_ADDTOOLW = (WM_USER + 50)
  Private Const TTM_DELTOOLA = (WM_USER + 5)
  Private Const TTM_DELTOOLW = (WM_USER + 51)
  #If FullFeatured Or Debuging Then
    Private Const TTM_GETDELAYTIME = (WM_USER + 21)
    Private Const TTM_GETMARGIN = (WM_USER + 27)
    Private Const TTM_GETMAXTIPWIDTH = (WM_USER + 25)
  #End If
  Private Const TTM_GETTEXTA = (WM_USER + 11)
  Private Const TTM_GETTEXTW = (WM_USER + 56)
  #If CustomDraw Or FullFeatured Or Debuging Then
    Private Const TTM_GETTIPBKCOLOR = (WM_USER + 22)
  #End If
  #If FullFeatured Or Debuging Then
    Private Const TTM_GETTIPTEXTCOLOR = (WM_USER + 23)
    Private Const TTM_GETTITLE = (WM_USER + 35)
  #End If
  Private Const TTM_POP = (WM_USER + 28)
  #If FullFeatured Or Debuging Then
    Private Const TTM_SETDELAYTIME = (WM_USER + 3)
    Private Const TTM_SETMARGIN = (WM_USER + 26)
    Private Const TTM_SETMAXTIPWIDTH = (WM_USER + 24)
    Private Const TTM_SETTIPBKCOLOR = (WM_USER + 19)
    Private Const TTM_SETTIPTEXTCOLOR = (WM_USER + 20)
    Private Const TTM_SETTITLEA = (WM_USER + 32)
    Private Const TTM_SETTITLEW = (WM_USER + 33)
    Private Const TTM_TRACKACTIVATE = (WM_USER + 17)
    Private Const TTM_TRACKPOSITION = (WM_USER + 18)
  #End If
  Private Const TTM_UPDATETIPTEXTA = (WM_USER + 12)
  Private Const TTM_UPDATETIPTEXTW = (WM_USER + 57)

  ' common controls notifications
  #If Debuging Then
    Private Const NM_CHAR = (NM_FIRST - 18)
    Private Const NM_CLICK = (NM_FIRST - 2)
  #End If
  #If CustomDraw Or Debuging Then
    Private Const NM_CUSTOMDRAW = (NM_FIRST - 12)
  #End If
  #If Debuging Then
    Private Const NM_DBLCLK = (NM_FIRST - 3)
    Private Const NM_HOVER = (NM_FIRST - 13)
    Private Const NM_KEYDOWN = (NM_FIRST - 15)
    Private Const NM_KILLFOCUS = (NM_FIRST - 8)
    Private Const NM_LDOWN = (NM_FIRST - 20)
    Private Const NM_NCHITTEST = (NM_FIRST - 14)
    Private Const NM_OUTOFMEMORY = (NM_FIRST - 1)
    Private Const NM_RCLICK = (NM_FIRST - 5)
    Private Const NM_RDBLCLK = (NM_FIRST - 6)
    Private Const NM_RDOWN = (NM_FIRST - 21)
    Private Const NM_RELEASEDCAPTURE = (NM_FIRST - 16)
    Private Const NM_RETURN = (NM_FIRST - 4)
    Private Const NM_SETCURSOR = (NM_FIRST - 17)
    Private Const NM_SETFOCUS = (NM_FIRST - 7)
    Private Const NM_THEMECHANGED = (NM_FIRST - 22)
    Private Const NM_TOOLTIPSCREATED = (NM_FIRST - 19)
  #End If

  ' tooltip notifications
  #If FullFeatured Or Debuging Then
    Private Const TTN_GETDISPINFOA = (TTN_FIRST - 0)
    Private Const TTN_GETDISPINFOW = (TTN_FIRST - 10)
    Private Const TTN_LINKCLICK = (TTN_FIRST - 3)
    Private Const TTN_NEEDTEXTA = TTN_GETDISPINFOA
    Private Const TTN_NEEDTEXTW = TTN_GETDISPINFOW
    Private Const TTN_POP = (TTN_FIRST - 2)
  #End If
  #If CustomDraw Or FullFeatured Or Debuging Then
    Private Const TTN_SHOW = (TTN_FIRST - 1)
  #End If

  ' tooltip flags
  #If Debuging Then
    Private Const TTF_ABSOLUTE = &H80
    Private Const TTF_RTLREADING = &H4
  #End If
  Private Const TTF_CENTERTIP = &H2
  Private Const TTF_DI_SETITEM = &H8000
  Private Const TTF_IDISHWND = &H1
  #If FullFeatured Or Debuging Then
    Private Const TTF_PARSELINKS = &H1000
  #End If
  Private Const TTF_SUBCLASS = &H10
  #If FullFeatured Or Debuging Then
    Private Const TTF_TRACK = &H20
  #End If
  Private Const TTF_TRANSPARENT = &H100

  ' constants for TTM_GETDELAYTIME
  #If Debuging Then
    Private Const TTDT_AUTOMATIC = 0
  #End If
  #If FullFeatured Or Debuging Then
    Private Const TTDT_AUTOPOP = 2
    Private Const TTDT_INITIAL = 3
    Private Const TTDT_RESHOW = 1
  #End If

  ' constants for TTM_SETTITLE
  #If Debuging Then
    Private Const TTI_ERROR = 3
    Private Const TTI_INFO = 1
    Private Const TTI_NONE = 0
    Private Const TTI_WARNING = 2
  #End If

  ' constants used with GetWindowLongPtr()
  #If Debuging Then
    Private Const DWL_DLGPROC = 4
    Private Const DWL_MSGRESULT = 0
    Private Const DWL_USER = 8
    Private Const DWLP_MSGRESULT = 0
    Private Const DWLP_DLGPROC = DWLP_MSGRESULT + 4 'LenB(LRESULT)
    Private Const DWLP_USER = DWLP_DLGPROC + 4 'LenB(DLGPROC)
    Private Const GWL_HINSTANCE = (-6)
    Private Const GWL_HWNDPARENT = (-8)
    Private Const GWL_ID = (-12)
    Private Const GWL_USERDATA = (-21)
    Private Const GWL_WNDPROC = (-4)
    Private Const GWLP_HINSTANCE = (-6)
    Private Const GWLP_HWNDPARENT = (-8)
    Private Const GWLP_ID = (-12)
    Private Const GWLP_USERDATA = (-21)
  #End If
  Private Const GWL_EXSTYLE = (-20)
  Private Const GWL_STYLE = (-16)
  Private Const GWLP_WNDPROC = (-4)

  ' constants used with ShowWindow()
  #If Debuging Then
    Private Const SW_FORCEMINIMIZE = 11
    Private Const SW_MAX = 11
    Private Const SW_MAXIMIZE = 3
    Private Const SW_MINIMIZE = 6
    Private Const SW_NORMAL = 1
    Private Const SW_RESTORE = 9
    Private Const SW_SHOW = 5
    Private Const SW_SHOWDEFAULT = 10
    Private Const SW_SHOWMAXIMIZED = 3
    Private Const SW_SHOWMINIMIZED = 2
    Private Const SW_SHOWMINNOACTIVE = 7
    Private Const SW_SHOWNA = 8
    Private Const SW_SHOWNOACTIVATE = 4
    Private Const SW_SHOWNORMAL = 1
  #End If
  Private Const SW_HIDE = 0


  #If CustomDraw Or Debuging Then
    Private Type BITMAP
      bmType As Long
      bmWidth As Long
      bmHeight As Long
      bmWidthBytes As Long
      bmPlanes As Integer
      bmBitsPixel As Integer
      bmBits As Long
    End Type
  #End If

  Private Type NMHDR
    hwndFrom As Long
    idfrom As Long
    code As Long
  End Type

  Private Type RECT
    Left As Long
    Top As Long
    Right As Long
    Bottom As Long
  End Type

  #If CustomDraw Or Debuging Then
    Private Type NMCUSTOMDRAW
      hdr As NMHDR
      dwDrawStage As Long
      hDC As Long
      rc As RECT
      dwItemSpec As Long
      uItemState As Long
      lItemlParam As Long
    End Type

    Private Type NMTTCUSTOMDRAW
      nmcd As NMCUSTOMDRAW
      uDrawFlags As Long
    End Type

    Private Type PAINTSTRUCT
      hDC As Long
      fErase As Long
      rcPaint As RECT
      fRestore As Long
      fIncUpdate As Long
      rgbReserved(32) As Byte
    End Type

    Private Type SIZE
      cx As Long
      cy As Long
    End Type
  #End If

  Private Type TOOLINFOA
    cbSize As Long
    uFlags As Long
    hWnd As Long
    uId As Long
    RECT As RECT
    hinst As Long
    lpszText As Long
    lParam As Long
    lpReserved As Long     ' needs Windows XP or higher
  End Type

  Private Type TOOLINFOW
    cbSize As Long
    uFlags As Long
    hWnd As Long
    uId As Long
    RECT As RECT
    hinst As Long
    lpszText As Long
    lParam As Long
    lpReserved As Long     ' needs Windows XP or higher
  End Type

  #If Debuging Then
    Private Type POINT
      x As Long
      y As Long
    End Type

    Private Type NMCHAR
      hdr As NMHDR
      ch As Long
      dwItemPrev As Long
      dwItemNext As Long
    End Type

    Private Type NMKEY
      hdr As NMHDR
      nVKey As Long
      uFlags As Long
    End Type

    Private Type NMMOUSE
      hdr As NMHDR
      dwItemSpec As Long
      dwItemData As Long
      pt As POINT
      dwHitInfo As POINT
    End Type

    Private Type NMOBJECTNOTIFY
      hdr As NMHDR
      IItem As Long
      pObject As Long
      hResult As Long
      dwFlags As Long
    End Type

    Private Type NMTOOLTIPSCREATED
      hdr As NMHDR
      hwndToolTips As Long
    End Type

    Private Type TT_HITTESTINFOA
      hWnd As Long
      pt As POINT
      ti As TOOLINFOA
    End Type

    Private Type TT_HITTESTINFOW
      hWnd As Long
      pt As POINT
      ti As TOOLINFOW
    End Type
  #End If

  Private Type INITCOMMONCONTROLSEXData
    dwSize As Long
    dwICC As Long
  End Type

  #If FullFeatured Or Debuging Then
    Private Type NMTTDISPINFOA
      hdr As NMHDR
      lpszText As Long
      szText(80) As Byte
      hinst As Long
      uFlags As Long
      lParam As Long
    End Type

    Private Type NMTTDISPINFOW
      hdr As NMHDR
      lpszText As Long
      szText(160) As Byte
      hinst As Long
      uFlags As Long
      lParam As Long
    End Type

    Private Type TTGETTITLE
      dwSize As Long
      uTitleBitmap As Long
      cch As Long
      pszTitle As Long
    End Type
  #End If


  Private bAttached As Boolean
  #If CustomDraw Or Debuging Then
    Private BMPSize As SIZE
  #End If
  #If CustomDraw Or FireEvents Or Debuging Then
    Private mASM() As Byte
  #End If
  Private oldWinMain As Long
  Private oldWinMainParent As Long
  #If CustomDraw Or Debuging Then
    Private rcContent As RECT
    Private TextSize As SIZE
  #End If

  #If FullFeatured Or Debuging Then
    Private propAlwaysTip As Boolean
    Private propAutoPopDelayTime As Integer
    Private propBackColor As OLE_COLOR
    Private propBalloonStyle As Boolean
  #End If
  #If CustomDraw Or Debuging Then
    Private propBitmapMarginX As Long
    Private propBitmapMarginY As Long
    Private propBitmapPosition As BitmapPositionConstants
    Private prophBitmap As Long
  #End If
  #If FullFeatured Or Debuging Then
    Private prophFont As Long
  #End If
  Private prophWnd As Long
  Private prophWndParent As Long
  #If FullFeatured Or Debuging Then
    Private propInitialDelayTime As Long
    Private propMargin As RECT
  #End If
  #If CustomDraw Or Debuging Then
    Private propMaskColor As OLE_COLOR
  #End If
  #If FullFeatured Or Debuging Then
    Private propMaxTipWidth As Long
    Private propNoAnimation As Boolean
    Private propNoFade As Boolean
    Private propReshowDelayTime As Long
    Private propShowCloseButton As Boolean
  #End If
  Private propText As String
  #If FullFeatured Or Debuging Then
    Private propTextColor As OLE_COLOR
    Private propTitleIcon As ToolTipTitleIconConstants
    Private propTitleText As String
  #End If
  Private propUnicode As Boolean
  #If CustomDraw Or Debuging Then
    Private propUseMaskColor As Boolean
  #End If


  #If CustomDraw Or Debuging Then
    Private Declare Function BeginPaint Lib "user32.dll" (ByVal hWnd As Long, Data As PAINTSTRUCT) As Long
    Private Declare Function BitBlt Lib "gdi32.dll" (ByVal hDC_Dst As Long, ByVal xDst As Long, ByVal yDst As Long, ByVal cxDst As Long, ByVal cyDst As Long, ByVal hDC_Src As Long, ByVal xSrc As Long, ByVal ySrc As Long, ByVal Mode As Long) As Long
  #End If
  #If CustomDraw Or FireEvents Or Debuging Then
    Private Declare Function CallWindowProcA Lib "user32.dll" (ByVal adr As Long, ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    Private Declare Function CallWindowProcW Lib "user32.dll" (ByVal adr As Long, ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  #End If
  Private Declare Sub CopyMemory Lib "kernel32.dll" Alias "RtlMoveMemory" (ByRef pDest As Any, ByRef pSrc As Any, ByVal SIZE As Long)
  #If CustomDraw Or Debuging Then
    Private Declare Function CreateBitmapAsLong Lib "gdi32.dll" Alias "CreateBitmap" (ByVal cx As Long, ByVal cy As Long, ByVal ClrPlanes As Long, ByVal BPP As Long, ByVal Bits As Long) As Long
    Private Declare Function CreateCompatibleDC Lib "gdi32.dll" (ByVal hDC As Long) As Long
    Private Declare Function CreateCompatibleBitmap Lib "gdi32.dll" (ByVal hDC As Long, ByVal Width As Long, ByVal Height As Long) As Long
    Private Declare Function CreateSolidBrush Lib "gdi32.dll" (ByVal Clr As Long) As Long
  #End If
  Private Declare Function CreateWindowExAAsLong Lib "user32.dll" Alias "CreateWindowExA" (ByVal StyleEx As Long, ByVal Klasse As String, ByVal Caption As String, ByVal Style As Long, ByVal x As Long, ByVal y As Long, ByVal Width As Long, ByVal Height As Long, ByVal hWnd_Parent As Long, ByVal hMenu As Long, ByVal hInstanz As Long, ByVal lParam As Long) As Long
  Private Declare Function CreateWindowExWAsLong Lib "user32.dll" Alias "CreateWindowExW" (ByVal StyleEx As Long, ByVal Klasse As String, ByVal Caption As String, ByVal Style As Long, ByVal x As Long, ByVal y As Long, ByVal Width As Long, ByVal Height As Long, ByVal hWnd_Parent As Long, ByVal hMenu As Long, ByVal hInstanz As Long, ByVal lParam As Long) As Long
  #If CustomDraw Or Debuging Then
    Private Declare Function DeleteDC Lib "gdi32.dll" (ByVal hDC As Long) As Long
    Private Declare Function DeleteObject Lib "gdi32.dll" (ByVal hObject As Long) As Long
  #End If
  Private Declare Function DestroyWindow Lib "user32.dll" (ByVal hWnd As Long) As Long
  #If CustomDraw Or Debuging Then
    Private Declare Function EndPaint Lib "user32.dll" (ByVal hWnd As Long, Data As PAINTSTRUCT) As Long
    Private Declare Function FillRect Lib "user32.dll" (ByVal hDC As Long, rc As RECT, ByVal hBrush As Long) As Long
    Private Declare Function GetClientRect Lib "user32.dll" (ByVal hWnd As Long, rc As RECT) As Long
    Private Declare Function GetObjectAPIA Lib "gdi32.dll" Alias "GetObjectA" (ByVal hObject As Long, ByVal Buffersize As Long, Data As Any) As Long
    Private Declare Function GetObjectAPIW Lib "gdi32.dll" Alias "GetObjectW" (ByVal hObject As Long, ByVal Buffersize As Long, Data As Any) As Long
  #End If
  Private Declare Function GetParent Lib "user32.dll" (ByVal hWnd As Long) As Long
  #If FullFeatured Or Debuging Then
    Private Declare Function GetWindowLongPtrA Lib "user32.dll" Alias "GetWindowLongA" (ByVal hWnd As Long, ByVal Prop As Long) As Long
    Private Declare Function GetWindowLongPtrW Lib "user32.dll" Alias "GetWindowLongW" (ByVal hWnd As Long, ByVal Prop As Long) As Long
  #End If
  Private Declare Function InitCommonControlsEx Lib "comctl32.dll" (Data As INITCOMMONCONTROLSEXData) As Long
  Private Declare Function IsWindowUnicode Lib "user32.dll" (ByVal hWnd As Long) As Long
  #If FullFeatured Or Debuging Then
    Private Declare Function lstrcpynAAsLong Lib "kernel32.dll" Alias "lstrcpynA" (ByVal pTarget As Long, ByVal Src As String, ByVal MaxLength As Long) As Long
    Private Declare Function lstrcpynWAsLong Lib "kernel32.dll" Alias "lstrcpynW" (ByVal pTarget As Long, ByVal Src As String, ByVal MaxLength As Long) As Long
    Private Declare Function OleTranslateColor Lib "olepro32.dll" (ByVal Clr As Long, ByVal hPal As Long, ret As Long) As Long
    Private Declare Function PostMessageA Lib "user32.dll" (ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    Private Declare Function PostMessageW Lib "user32.dll" (ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  #End If
  #If CustomDraw Or Debuging Then
    Private Declare Function SelectObject Lib "gdi32.dll" (ByVal hDC As Long, ByVal hObject As Long) As Long
  #End If
  Private Declare Function SendMessageA Lib "user32.dll" (ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, lParam As Any) As Long
  Private Declare Function SendMessageW Lib "user32.dll" (ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, lParam As Any) As Long
  Private Declare Function SendMessageAAsLong Lib "user32.dll" Alias "SendMessageA" (ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Private Declare Function SendMessageWAsLong Lib "user32.dll" Alias "SendMessageW" (ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  #If CustomDraw Or Debuging Then
    Private Declare Function SetBkColor Lib "gdi32.dll" (ByVal hDC As Long, ByVal Clr As Long) As Long
  #End If
  Private Declare Function SetParent Lib "user32.dll" (ByVal hWndChild As Long, ByVal hWndNewParent As Long) As Long
  Private Declare Function SetWindowLongPtrA Lib "user32.dll" Alias "SetWindowLongA" (ByVal hWnd As Long, ByVal Prop As Long, ByVal NewProp As Long) As Long
  Private Declare Function SetWindowLongPtrW Lib "user32.dll" Alias "SetWindowLongW" (ByVal hWnd As Long, ByVal Prop As Long, ByVal NewProp As Long) As Long
  #If CustomDraw Or Debuging Then
    Private Declare Function SetWindowPos Lib "user32.dll" (ByVal hWnd As Long, ByVal hWndInsertAfter As Long, ByVal x As Long, ByVal y As Long, ByVal cx As Long, ByVal cy As Long, ByVal Flags As Long) As Long
  #End If
  Private Declare Function ShowWindow Lib "user32.dll" (ByVal hWnd As Long, ByVal Flags As Long) As Long


  #If FullFeatured Or Debuging Then
    Public Event BeforeHide()
    Public Event BeforeShow()
    Public Event ClickedLink()
    Public Event NeedText(ByRef TextToDisplay As String, ByRef StoreData As Boolean)
    Public Event ProcessMessage(ByVal oldWinMain As Long, ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long, ByVal callOldWinMain As Boolean, ByVal ValueToReturn As Long)
  #End If


Private Sub Class_Initialize()
  Dim Data As INITCOMMONCONTROLSEXData
  Dim ptr As Long

  Data.dwICC = CommonControlsConsts.ICC_BAR_CLASSES
  Data.dwSize = LenB(Data)
  InitCommonControlsEx Data

  #If CustomDraw Or FireEvents Or Debuging Then
    ' this is a trick to simulate AddressOf for methods that belong to an object
    ' get the VTable-Ptr
    CopyMemory ptr, ByVal ObjPtr(Me), LenB(ptr)
    ' get the pointer to the first public procedure (must be our WinMain)
    CopyMemory ptr, ByVal ptr + &H1C, LenB(ptr)
    ReDim mASM(0 To 42) As Byte     ' bytes 38 to 42 will receive the return value of our WinMain
    ' lay the frame pointer onto the stack (what for?)
    mASM(0) = &H55     ' PUSH EBP
    ' get the stack pointer into EBP
    CopyMemory mASM(1), &H68EC8B, 3     ' MOV EBP, ESP
    ' lay the offset of the return value's address onto the stack
    CopyMemory mASM(4), VarPtr(mASM(39)), 4   ' PUSH dword ptr [offset of return address]

    ' lay the addresses of the WinMain's arguments onto the stack
    ' 4th argument
    CopyMemory mASM(8), &H1475FF, 3     ' PUSH dword ptr [EBP+14h]
    ' 3rd argument
    CopyMemory mASM(11), &H1075FF, 3     ' PUSH dword ptr [EBP+10h]
    ' 2nd argument
    CopyMemory mASM(14), &HC75FF, 3     ' PUSH dword ptr [EBP+0Ch]
    ' 1st argument
    CopyMemory mASM(17), &H875FF, 3     ' PUSH dword ptr [EBP+08h]

    ' lay the current object's pointer onto the stack
    mASM(20) = &H68     ' PUSH 32-bit-immediate
    CopyMemory mASM(21), ObjPtr(Me), 4     ' the immediate
    ' call our WinMain
    mASM(25) = &HE8     ' CALL 32-bit relative code address
    CopyMemory mASM(26), ptr - VarPtr(mASM(30)), 4    ' the relative address
    ' move the WinMain's return value into EAX
    mASM(30) = &HA1     ' MOV EAX, dword ptr [offset of return address]
    CopyMemory mASM(31), VarPtr(mASM(39)), 4     ' the offset
    ' leave our WinMain
    mASM(35) = &HC9     ' LEAVE
    ' return from our WinMain
    mASM(36) = &HC2     ' RET 10h     (seems to be some offset - it's the size of all arguments together)
    CopyMemory mASM(37), &H10, 2    ' can't do the "RET 10h" in 1 line - probably alignment problems
  #End If

  #If FullFeatured Or Debuging Then
    propAlwaysTip = True
    propAutoPopDelayTime = 5000
    propBackColor = SystemColorConstants.vbInfoBackground
    propBalloonStyle = False
  #End If
  #If CustomDraw Or Debuging Then
    propBitmapMarginX = 7
    propBitmapMarginY = 7
    propBitmapPosition = BitmapPositionConstants.bpLeft
  #End If
  #If FullFeatured Or Debuging Then
    propInitialDelayTime = 500
  #End If
  #If CustomDraw Or Debuging Then
    propMaskColor = ColorConstants.vbMagenta
  #End If
  #If FullFeatured Or Debuging Then
    propMaxTipWidth = 300
    propNoAnimation = False
    propNoFade = False
    propReshowDelayTime = 100
    propShowCloseButton = False
    propTextColor = SystemColorConstants.vbInfoText
    propTitleIcon = ToolTipTitleIconConstants.tttiNone
    propTitleText = ""
  #End If
  propUnicode = True
  #If CustomDraw Or Debuging Then
    propUseMaskColor = True
  #End If
End Sub

Private Sub Class_Terminate()
  If bAttached Then
    Me.Detach
  Else
    Me.Destroy
  End If
End Sub


#If CustomDraw Or FireEvents Or Debuging Then
  ' This must be the first public method!!
  ' this is *our* WinMain for the tooltip and its parent
  Public Function WinMain(ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    Dim ret As Long

    Select Case hWnd
      Case Me.hWnd
        ret = WinMain_ToolTip(hWnd, Msg, wParam, lParam)
      Case Me.hWndParent
        ret = WinMain_Parent(hWnd, Msg, wParam, lParam)
    End Select

    WinMain = ret
  End Function
#End If


#If FullFeatured Or Debuging Then
  Public Property Get AlwaysTip() As Boolean
    Dim Style As Long

    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      propAlwaysTip = ((Style And ToolTipStyleConsts.TTS_ALWAYSTIP) = ToolTipStyleConsts.TTS_ALWAYSTIP)
    End If
    AlwaysTip = propAlwaysTip
  End Property

  Public Property Let AlwaysTip(ByVal newVal As Boolean)
    Dim Style As Long

    propAlwaysTip = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      If propAlwaysTip Then
        Style = Style Or ToolTipStyleConsts.TTS_ALWAYSTIP
      Else
        Style = Style And Not ToolTipStyleConsts.TTS_ALWAYSTIP
      End If
      If Me.Unicode Then
        SetWindowLongPtrW Me.hWnd, GWL_STYLE, Style
      Else
        SetWindowLongPtrA Me.hWnd, GWL_STYLE, Style
      End If
    End If
  End Property

  Public Property Get AutoPopDelayTime() As Integer
    If Me.hWnd Then
      If Me.Unicode Then
        propAutoPopDelayTime = SendMessageWAsLong(Me.hWnd, TTM_GETDELAYTIME, TTDT_AUTOPOP, 0)
      Else
        propAutoPopDelayTime = SendMessageAAsLong(Me.hWnd, TTM_GETDELAYTIME, TTDT_AUTOPOP, 0)
      End If
    End If
    AutoPopDelayTime = propAutoPopDelayTime
  End Property

  Public Property Let AutoPopDelayTime(ByVal newVal As Integer)
    propAutoPopDelayTime = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageWAsLong Me.hWnd, TTM_SETDELAYTIME, TTDT_AUTOPOP, MAKELONG(propAutoPopDelayTime, 0)
      Else
        SendMessageAAsLong Me.hWnd, TTM_SETDELAYTIME, TTDT_AUTOPOP, MAKELONG(propAutoPopDelayTime, 0)
      End If
    End If
  End Property

  Public Property Get BackColor() As OLE_COLOR
    If Me.hWnd Then
      If Me.Unicode Then
        propBackColor = SendMessageWAsLong(Me.hWnd, TTM_GETTIPBKCOLOR, 0, 0)
      Else
        propBackColor = SendMessageAAsLong(Me.hWnd, TTM_GETTIPBKCOLOR, 0, 0)
      End If
    End If
    BackColor = propBackColor
  End Property

  Public Property Let BackColor(ByVal newVal As OLE_COLOR)
    propBackColor = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        PostMessageW Me.hWnd, TTM_SETTIPBKCOLOR, TranslateColor(propBackColor), 0
      Else
        PostMessageA Me.hWnd, TTM_SETTIPBKCOLOR, TranslateColor(propBackColor), 0
      End If
    End If
  End Property

  Public Property Get BalloonStyle() As Boolean
    Dim Style As Long

    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      propBalloonStyle = ((Style And ToolTipStyleConsts.TTS_BALLOON) = ToolTipStyleConsts.TTS_BALLOON)
    End If
    BalloonStyle = propBalloonStyle
  End Property

  Public Property Let BalloonStyle(ByVal newVal As Boolean)
    Dim Style As Long

    propBalloonStyle = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      If propBalloonStyle Then
        Style = Style Or ToolTipStyleConsts.TTS_BALLOON
        Style = Style And Not WindowStyleConsts.WS_BORDER
      Else
        Style = Style And Not ToolTipStyleConsts.TTS_BALLOON
        Style = Style Or WindowStyleConsts.WS_BORDER
      End If
      If Me.Unicode Then
        SetWindowLongPtrW Me.hWnd, GWL_STYLE, Style
      Else
        SetWindowLongPtrA Me.hWnd, GWL_STYLE, Style
      End If
    End If
  End Property
#End If

#If CustomDraw Or Debuging Then
  Public Property Get BitmapMarginX() As Long
    BitmapMarginX = propBitmapMarginX
  End Property

  Public Property Let BitmapMarginX(ByVal newVal As Long)
    propBitmapMarginX = newVal
  End Property

  Public Property Get BitmapMarginY() As Long
    BitmapMarginY = propBitmapMarginY
  End Property

  Public Property Let BitmapMarginY(ByVal newVal As Long)
    propBitmapMarginY = newVal
  End Property

  Public Property Get BitmapPosition() As BitmapPositionConstants
    BitmapPosition = propBitmapPosition
  End Property

  Public Property Let BitmapPosition(ByVal newVal As BitmapPositionConstants)
    propBitmapPosition = newVal
  End Property

  Public Property Get hBitmap() As Long
    hBitmap = prophBitmap
  End Property

  Public Property Let hBitmap(ByVal newVal As Long)
    Dim BMPData As BITMAP

    prophBitmap = newVal

    If Me.Unicode Then
      GetObjectAPIW prophBitmap, LenB(BMPData), BMPData
    Else
      GetObjectAPIA prophBitmap, LenB(BMPData), BMPData
    End If
    BMPSize.cx = BMPData.bmWidth
    BMPSize.cy = BMPData.bmHeight
  End Property
#End If

#If FullFeatured Or Debuging Then
  Public Property Get hFont() As Long
    Const WM_GETFONT = &H31

    If Me.hWnd Then
      If Me.Unicode Then
        prophFont = SendMessageWAsLong(Me.hWnd, WM_GETFONT, 0, 0)
      Else
        prophFont = SendMessageAAsLong(Me.hWnd, WM_GETFONT, 0, 0)
      End If
    End If
    hFont = prophFont
  End Property

  Public Property Let hFont(ByVal newVal As Long)
    Const WM_SETFONT = &H30

    prophFont = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        PostMessageW Me.hWnd, WM_SETFONT, prophFont, 1
      Else
        PostMessageA Me.hWnd, WM_SETFONT, prophFont, 1
      End If
    End If
  End Property
#End If

Public Property Get hWnd() As Long
  hWnd = prophWnd
End Property

Public Property Get hWndParent() As Long
  If Me.hWnd Then prophWndParent = GetParent(Me.hWnd)
  hWndParent = prophWndParent
End Property

Public Property Let hWndParent(ByVal newVal As Long)
  UnSubclassParent
  prophWndParent = newVal
  SubclassParent
  If Me.hWnd Then SetParent Me.hWnd, prophWndParent
End Property

#If FullFeatured Or Debuging Then
  Public Property Get InitialDelayTime() As Integer
    If Me.hWnd Then
      If Me.Unicode Then
        propInitialDelayTime = SendMessageWAsLong(Me.hWnd, TTM_GETDELAYTIME, TTDT_INITIAL, 0)
      Else
        propInitialDelayTime = SendMessageAAsLong(Me.hWnd, TTM_GETDELAYTIME, TTDT_INITIAL, 0)
      End If
    End If
    InitialDelayTime = propInitialDelayTime
  End Property

  Public Property Let InitialDelayTime(ByVal newVal As Integer)
    propInitialDelayTime = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageWAsLong Me.hWnd, TTM_SETDELAYTIME, TTDT_INITIAL, MAKELONG(propInitialDelayTime, 0)
      Else
        SendMessageAAsLong Me.hWnd, TTM_SETDELAYTIME, TTDT_INITIAL, MAKELONG(propInitialDelayTime, 0)
      End If
    End If
  End Property

  Public Property Get MarginBottom() As Long
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageW Me.hWnd, TTM_GETMARGIN, 0, propMargin
      Else
        SendMessageA Me.hWnd, TTM_GETMARGIN, 0, propMargin
      End If
    End If
    MarginBottom = propMargin.Bottom
  End Property

  Public Property Let MarginBottom(ByVal newVal As Long)
    propMargin.Bottom = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageW Me.hWnd, TTM_SETMARGIN, 0, propMargin
      Else
        SendMessageA Me.hWnd, TTM_SETMARGIN, 0, propMargin
      End If
    End If
  End Property

  Public Property Get MarginLeft() As Long
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageW Me.hWnd, TTM_GETMARGIN, 0, propMargin
      Else
        SendMessageA Me.hWnd, TTM_GETMARGIN, 0, propMargin
      End If
    End If
    MarginLeft = propMargin.Left
  End Property

  Public Property Let MarginLeft(ByVal newVal As Long)
    propMargin.Left = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageW Me.hWnd, TTM_SETMARGIN, 0, propMargin
      Else
        SendMessageA Me.hWnd, TTM_SETMARGIN, 0, propMargin
      End If
    End If
  End Property

  Public Property Get MarginRight() As Long
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageW Me.hWnd, TTM_GETMARGIN, 0, propMargin
      Else
        SendMessageA Me.hWnd, TTM_GETMARGIN, 0, propMargin
      End If
    End If
    MarginRight = propMargin.Right
  End Property

  Public Property Let MarginRight(ByVal newVal As Long)
    propMargin.Right = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageW Me.hWnd, TTM_SETMARGIN, 0, propMargin
      Else
        SendMessageA Me.hWnd, TTM_SETMARGIN, 0, propMargin
      End If
    End If
  End Property

  Public Property Get MarginTop() As Long
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageW Me.hWnd, TTM_GETMARGIN, 0, propMargin
      Else
        SendMessageA Me.hWnd, TTM_GETMARGIN, 0, propMargin
      End If
    End If
    MarginTop = propMargin.Top
  End Property

  Public Property Let MarginTop(ByVal newVal As Long)
    propMargin.Top = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageW Me.hWnd, TTM_SETMARGIN, 0, propMargin
      Else
        SendMessageA Me.hWnd, TTM_SETMARGIN, 0, propMargin
      End If
    End If
  End Property
#End If

#If CustomDraw Or Debuging Then
  Public Property Get MaskColor() As OLE_COLOR
    MaskColor = propMaskColor
  End Property

  Public Property Let MaskColor(ByVal newVal As OLE_COLOR)
    propMaskColor = newVal
  End Property
#End If

#If FullFeatured Or Debuging Then
  Public Property Get MaxTipWidth() As Long
    If Me.hWnd Then
      If Me.Unicode Then
        propMaxTipWidth = SendMessageWAsLong(Me.hWnd, TTM_GETMAXTIPWIDTH, 0, 0)
      Else
        propMaxTipWidth = SendMessageAAsLong(Me.hWnd, TTM_GETMAXTIPWIDTH, 0, 0)
      End If
    End If
    MaxTipWidth = propMaxTipWidth
  End Property

  Public Property Let MaxTipWidth(ByVal newVal As Long)
    propMaxTipWidth = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageWAsLong Me.hWnd, TTM_SETMAXTIPWIDTH, 0, propMaxTipWidth
      Else
        SendMessageAAsLong Me.hWnd, TTM_SETMAXTIPWIDTH, 0, propMaxTipWidth
      End If
    End If
  End Property

  Public Property Get NoAnimation() As Boolean
    Dim Style As Long

    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      propNoAnimation = ((Style And ToolTipStyleConsts.TTS_NOANIMATE) = ToolTipStyleConsts.TTS_NOANIMATE)
    End If
    NoAnimation = propNoAnimation
  End Property

  Public Property Let NoAnimation(ByVal newVal As Boolean)
    Dim Style As Long

    propNoAnimation = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      If propNoAnimation Then
        Style = Style Or ToolTipStyleConsts.TTS_NOANIMATE
      Else
        Style = Style And Not ToolTipStyleConsts.TTS_NOANIMATE
      End If
      If Me.Unicode Then
        SetWindowLongPtrW Me.hWnd, GWL_STYLE, Style
      Else
        SetWindowLongPtrA Me.hWnd, GWL_STYLE, Style
      End If
    End If
  End Property

  Public Property Get NoFade() As Boolean
    Dim Style As Long

    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      propNoFade = ((Style And ToolTipStyleConsts.TTS_NOFADE) = ToolTipStyleConsts.TTS_NOFADE)
    End If
    NoFade = propNoFade
  End Property

  Public Property Let NoFade(ByVal newVal As Boolean)
    Dim Style As Long

    propNoFade = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      If propNoFade Then
        Style = Style Or ToolTipStyleConsts.TTS_NOFADE
      Else
        Style = Style And Not ToolTipStyleConsts.TTS_NOFADE
      End If
      If Me.Unicode Then
        SetWindowLongPtrW Me.hWnd, GWL_STYLE, Style
      Else
        SetWindowLongPtrA Me.hWnd, GWL_STYLE, Style
      End If
    End If
  End Property

  Public Property Get ReshowDelayTime() As Integer
    If Me.hWnd Then
      If Me.Unicode Then
        propReshowDelayTime = SendMessageWAsLong(Me.hWnd, TTM_GETDELAYTIME, TTDT_RESHOW, 0)
      Else
        propReshowDelayTime = SendMessageAAsLong(Me.hWnd, TTM_GETDELAYTIME, TTDT_RESHOW, 0)
      End If
    End If
    ReshowDelayTime = propReshowDelayTime
  End Property

  Public Property Let ReshowDelayTime(ByVal newVal As Integer)
    propReshowDelayTime = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageWAsLong Me.hWnd, TTM_SETDELAYTIME, TTDT_RESHOW, MAKELONG(propReshowDelayTime, 0)
      Else
        SendMessageAAsLong Me.hWnd, TTM_SETDELAYTIME, TTDT_RESHOW, MAKELONG(propReshowDelayTime, 0)
      End If
    End If
  End Property

  Public Property Get ShowCloseButton() As Boolean
    Dim Style As Long

    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      propShowCloseButton = ((Style And ToolTipStyleConsts.TTS_CLOSE) = ToolTipStyleConsts.TTS_CLOSE)
    End If
    ShowCloseButton = propShowCloseButton
  End Property

  Public Property Let ShowCloseButton(ByVal newVal As Boolean)
    Dim Style As Long

    propShowCloseButton = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        Style = GetWindowLongPtrW(Me.hWnd, GWL_STYLE)
      Else
        Style = GetWindowLongPtrA(Me.hWnd, GWL_STYLE)
      End If
      If propShowCloseButton Then
        Style = Style Or ToolTipStyleConsts.TTS_CLOSE
      Else
        Style = Style And Not ToolTipStyleConsts.TTS_CLOSE
      End If
      If Me.Unicode Then
        SetWindowLongPtrW Me.hWnd, GWL_STYLE, Style
      Else
        SetWindowLongPtrA Me.hWnd, GWL_STYLE, Style
      End If
    End If
  End Property
#End If

Public Property Get Text(ByVal hWndTool As Long) As String
  Dim DataA As TOOLINFOA
  Dim DataW As TOOLINFOW

  If Me.hWnd Then
    If Me.Unicode Then
      With DataW
        .cbSize = LenB(DataW) - LenB(.lpReserved)
        .hinst = App.hInstance
        .hWnd = Me.hWndParent
        propText = Space(2048)
        .lpszText = StrPtr(propText)
        .uId = hWndTool
        .uFlags = TTF_IDISHWND
      End With
      SendMessageW Me.hWnd, TTM_GETTEXTW, 0, DataW
      propText = removeNull(propText)
    Else
      With DataA
        .cbSize = LenB(DataA) - LenB(.lpReserved)
        .hinst = App.hInstance
        .hWnd = Me.hWndParent
        propText = Space(2048)
        .lpszText = StrPtr(propText)
        .uId = hWndTool
        .uFlags = TTF_IDISHWND
      End With
      SendMessageA Me.hWnd, TTM_GETTEXTA, 0, DataA
      propText = removeNull(StrConv(propText, VbStrConv.vbUnicode))
    End If
  End If
  Text = propText
End Property

Public Property Let Text(ByVal hWndTool As Long, ByVal newVal As String)
  Dim DataA As TOOLINFOA
  Dim DataW As TOOLINFOW

  propText = newVal
  If Me.hWnd Then
    If Me.Unicode Then
      With DataW
        .cbSize = LenB(DataW) - LenB(.lpReserved)
        .hinst = App.hInstance
        .hWnd = Me.hWndParent
        .lpszText = StrPtr(propText)
        .uId = hWndTool
        .uFlags = TTF_IDISHWND
      End With
      SendMessageW Me.hWnd, TTM_UPDATETIPTEXTW, 0, DataW
    Else
      With DataA
        .cbSize = LenB(DataA) - LenB(.lpReserved)
        .hinst = App.hInstance
        .hWnd = Me.hWndParent
        .lpszText = StrPtr(StrConv(propText, VbStrConv.vbFromUnicode))
        .uId = hWndTool
        .uFlags = TTF_IDISHWND
      End With
      SendMessageA Me.hWnd, TTM_UPDATETIPTEXTA, 0, DataA
    End If
  End If
End Property

#If FullFeatured Or Debuging Then
  Public Property Get TextColor() As OLE_COLOR
    If Me.hWnd Then
      If Me.Unicode Then
        propTextColor = SendMessageWAsLong(Me.hWnd, TTM_GETTIPTEXTCOLOR, 0, 0)
      Else
        propTextColor = SendMessageAAsLong(Me.hWnd, TTM_GETTIPTEXTCOLOR, 0, 0)
      End If
    End If
    TextColor = propTextColor
  End Property

  Public Property Let TextColor(ByVal newVal As OLE_COLOR)
    propTextColor = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        PostMessageW Me.hWnd, TTM_SETTIPTEXTCOLOR, TranslateColor(propTextColor), 0
      Else
        PostMessageA Me.hWnd, TTM_SETTIPTEXTCOLOR, TranslateColor(propTextColor), 0
      End If
    End If
  End Property

  Public Property Get TitleIcon() As ToolTipTitleIconConstants
    Dim Data As TTGETTITLE
    Dim str As String

    If Me.hWnd Then
      With Data
        .dwSize = LenB(Data)
        str = Space(1024)
        .pszTitle = StrPtr(str)
        .cch = Len(str)

        If Me.Unicode Then
          SendMessageW Me.hWnd, TTM_GETTITLE, 0, Data
        Else
          SendMessageA Me.hWnd, TTM_GETTITLE, 0, Data
        End If
        propTitleIcon = .uTitleBitmap
      End With
    End If
    TitleIcon = propTitleIcon
  End Property

  Public Property Let TitleIcon(ByVal newVal As ToolTipTitleIconConstants)
    propTitleIcon = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageWAsLong Me.hWnd, TTM_SETTITLEW, propTitleIcon, StrPtr(propTitleText)
      Else
        SendMessageAAsLong Me.hWnd, TTM_SETTITLEA, propTitleIcon, StrPtr(StrConv(propTitleText, VbStrConv.vbFromUnicode))
      End If
    End If
  End Property

  Public Property Get TitleText() As String
    Dim Data As TTGETTITLE
    Dim str As String

    If Me.hWnd Then
      With Data
        .dwSize = LenB(Data)
        str = Space(1024)
        .pszTitle = StrPtr(str)
        .cch = Len(str)

        If Me.Unicode Then
          SendMessageW Me.hWnd, TTM_GETTITLE, 0, Data
        Else
          SendMessageA Me.hWnd, TTM_GETTITLE, 0, Data
        End If
        propTitleText = removeNull(str)
      End With
    End If
    TitleText = propTitleText
  End Property

  Public Property Let TitleText(ByVal newVal As String)
    propTitleText = newVal
    If Me.hWnd Then
      If Me.Unicode Then
        SendMessageWAsLong Me.hWnd, TTM_SETTITLEW, propTitleIcon, StrPtr(propTitleText)
      Else
        SendMessageAAsLong Me.hWnd, TTM_SETTITLEA, propTitleIcon, StrPtr(StrConv(propTitleText, VbStrConv.vbFromUnicode))
      End If
    End If
  End Property
#End If

Public Property Get Unicode() As Boolean
  If Me.hWnd Then
    propUnicode = IsWindowUnicode(Me.hWnd)
  End If

  Unicode = propUnicode
End Property

Public Property Let Unicode(ByVal newVal As Boolean)
  propUnicode = newVal
  #If FullFeatured Or Debuging Then
    If Me.hWnd Then Me.Create Me.hWndParent, prophFont
  #Else
    If Me.hWnd Then Me.Create Me.hWndParent
  #End If
End Property

#If CustomDraw Or Debuging Then
  Public Property Get UseMaskColor() As Boolean
    UseMaskColor = propUseMaskColor
  End Property

  Public Property Let UseMaskColor(ByVal newVal As Boolean)
    propUseMaskColor = newVal
  End Property
#End If


Public Sub Activate()
  If Me.hWnd Then
    If Me.Unicode Then
      SendMessageWAsLong Me.hWnd, TTM_ACTIVATE, 1, 0
    Else
      SendMessageAAsLong Me.hWnd, TTM_ACTIVATE, 1, 0
    End If
  End If
End Sub

#If FullFeatured Or Debuging Then
  Public Sub AddTool(ByVal hWndTool As Long, ByVal TextToDisplay As String, Optional ByVal SubClassTool As Boolean = True, Optional ByVal Transparent As Boolean = True, Optional ByVal AdjustToTool As Boolean = True, Optional ByVal ParseLinks As Boolean = True, Optional ByVal Track As Boolean = False)
#Else
  Public Sub AddTool(ByVal hWndTool As Long, ByVal TextToDisplay As String, Optional ByVal SubClassTool As Boolean = True, Optional ByVal Transparent As Boolean = True, Optional ByVal AdjustToTool As Boolean = True)
#End If
  Dim DataA As TOOLINFOA
  Dim DataW As TOOLINFOW

  If Me.hWnd Then
    If Me.Unicode Then
      With DataW
        .cbSize = LenB(DataW) - LenB(.lpReserved)
        .hinst = App.hInstance
        .hWnd = Me.hWndParent
        #If FullFeatured Or Debuging Then
          If TextToDisplay = "" Then
            .lpszText = LPSTR_TEXTCALLBACKW
          Else
            .lpszText = StrPtr(TextToDisplay)
          End If
        #Else
          .lpszText = StrPtr(TextToDisplay)
        #End If
        .uId = hWndTool
        .uFlags = TTF_IDISHWND
        If AdjustToTool Then .uFlags = .uFlags Or TTF_CENTERTIP
        #If FullFeatured Or Debuging Then
          If ParseLinks Then .uFlags = .uFlags Or TTF_PARSELINKS
        #End If
        If SubClassTool Then .uFlags = .uFlags Or TTF_SUBCLASS
        #If FullFeatured Or Debuging Then
          If Track Then .uFlags = .uFlags Or TTF_TRACK
        #End If
        If Transparent Then .uFlags = .uFlags Or TTF_TRANSPARENT
      End With
      SendMessageW Me.hWnd, TTM_ADDTOOLW, 0, DataW
    Else
      With DataA
        .cbSize = LenB(DataA) - LenB(.lpReserved)
        .hinst = App.hInstance
        .hWnd = Me.hWndParent
        #If FullFeatured Or Debuging Then
          If TextToDisplay = "" Then
            .lpszText = LPSTR_TEXTCALLBACKA
          Else
            .lpszText = StrPtr(StrConv(TextToDisplay, VbStrConv.vbFromUnicode))
          End If
        #Else
          .lpszText = StrPtr(StrConv(TextToDisplay, VbStrConv.vbFromUnicode))
        #End If
        .uId = hWndTool
        .uFlags = TTF_IDISHWND
        If AdjustToTool Then .uFlags = .uFlags Or TTF_CENTERTIP
        #If FullFeatured Or Debuging Then
          If ParseLinks Then .uFlags = .uFlags Or TTF_PARSELINKS
        #End If
        If SubClassTool Then .uFlags = .uFlags Or TTF_SUBCLASS
        #If FullFeatured Or Debuging Then
          If Track Then .uFlags = .uFlags Or TTF_TRACK
        #End If
        If Transparent Then .uFlags = .uFlags Or TTF_TRANSPARENT
      End With
      SendMessageA Me.hWnd, TTM_ADDTOOLA, 0, DataA
    End If
  End If
End Sub

' attaches this object to an existing tooltip control
Public Sub Attach(ByVal hWndToolTip As Long)
  Me.Destroy
  prophWnd = hWndToolTip
  If Me.hWnd Then
    Subclass
    SubclassParent
  End If
  bAttached = True
End Sub

' creates a tooltip
#If FullFeatured Or Debuging Then
  Public Sub Create(Optional ByVal hWndParent As Long = -1, Optional ByVal hFont As Long = -1)
#Else
  Public Sub Create(Optional ByVal hWndParent As Long = -1)
#End If
  Const CW_USEDEFAULT = &H80000000
  #If FullFeatured Or Debuging Then
    Const WM_GETFONT = &H31
    Const WM_SETFONT = &H30
  #End If
  Dim ExStyle As ExWindowStyleConsts
  Dim Style As WindowStyleConsts
  Dim TTStyle As ToolTipStyleConsts

  ' you'll never know...
  Me.Destroy
  If Me.hWnd = 0 Then
    If hWndParent <> -1 Then prophWndParent = hWndParent
    #If FullFeatured Or Debuging Then
      If hFont <> -1 Then
        Me.hFont = hFont
      ElseIf Me.hFont = -1 Then
        If Me.Unicode Then
          Me.hFont = SendMessageWAsLong(Me.hWndParent, WM_GETFONT, 0, 0)
        Else
          Me.hFont = SendMessageAAsLong(Me.hWndParent, WM_GETFONT, 0, 0)
        End If
      End If
    #End If

    ExStyle = ExWindowStyleConsts.WS_EX_LEFT Or ExWindowStyleConsts.WS_EX_LTRREADING 'Or ExWindowStyleConsts.WS_EX_TOOLWINDOW
    Style = WindowStyleConsts.WS_OVERLAPPED 'WindowStyleConsts.WS_POPUP Or WindowStyleConsts.WS_VISIBLE
    TTStyle = ToolTipStyleConsts.TTS_NOPREFIX
    #If FullFeatured Or Debuging Then
      If propAlwaysTip Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_ALWAYSTIP
      If propBalloonStyle Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_BALLOON
      If propNoAnimation Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_NOANIMATE
      If propNoFade Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_NOFADE
      If propShowCloseButton Then TTStyle = TTStyle Or ToolTipStyleConsts.TTS_CLOSE
    #End If

    ' create the window
    If Me.Unicode Then
      prophWnd = CreateWindowExWAsLong(ExStyle, StrConv(WC_TOOLTIP, VbStrConv.vbUnicode), "", Style Or TTStyle, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, Me.hWndParent, 0, App.hInstance, 0)
    Else
      prophWnd = CreateWindowExAAsLong(ExStyle, WC_TOOLTIP, "", Style Or TTStyle, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, Me.hWndParent, 0, App.hInstance, 0)
    End If
    If Me.hWnd = 0 Then GoTo MyError
  End If
  Subclass
  SubclassParent

  #If FullFeatured Or Debuging Then
    ' set the properties...
    If Me.Unicode Then
      PostMessageW Me.hWnd, TTM_SETMAXTIPWIDTH, 0, propMaxTipWidth
      PostMessageW Me.hWnd, TTM_SETDELAYTIME, TTDT_AUTOPOP, MAKELONG(propAutoPopDelayTime, 0)
      PostMessageW Me.hWnd, TTM_SETDELAYTIME, TTDT_INITIAL, MAKELONG(propInitialDelayTime, 0)
      PostMessageW Me.hWnd, TTM_SETDELAYTIME, TTDT_RESHOW, MAKELONG(propReshowDelayTime, 0)
      SendMessageWAsLong Me.hWnd, TTM_SETTITLEW, propTitleIcon, StrPtr(propTitleText)
    Else
      PostMessageA Me.hWnd, TTM_SETMAXTIPWIDTH, 0, propMaxTipWidth
      PostMessageA Me.hWnd, TTM_SETDELAYTIME, TTDT_AUTOPOP, MAKELONG(propAutoPopDelayTime, 0)
      PostMessageA Me.hWnd, TTM_SETDELAYTIME, TTDT_INITIAL, MAKELONG(propInitialDelayTime, 0)
      PostMessageA Me.hWnd, TTM_SETDELAYTIME, TTDT_RESHOW, MAKELONG(propReshowDelayTime, 0)
      SendMessageAAsLong Me.hWnd, TTM_SETTITLEA, propTitleIcon, StrPtr(StrConv(propTitleText, VbStrConv.vbFromUnicode))
    End If

    ' ...and colors...
    If Me.Unicode Then
      PostMessageW Me.hWnd, TTM_SETTIPBKCOLOR, TranslateColor(propBackColor), 0
      PostMessageW Me.hWnd, TTM_SETTIPTEXTCOLOR, TranslateColor(propTextColor), 0
    Else
      PostMessageA Me.hWnd, TTM_SETTIPBKCOLOR, TranslateColor(propBackColor), 0
      PostMessageA Me.hWnd, TTM_SETTIPTEXTCOLOR, TranslateColor(propTextColor), 0
    End If

    ' ...and fonts
    If Me.Unicode Then
      PostMessageW Me.hWnd, WM_SETFONT, prophFont, 1
    Else
      PostMessageA Me.hWnd, WM_SETFONT, prophFont, 1
    End If
  #End If
  Exit Sub

MyError:
End Sub

Public Sub Deactivate()
  If Me.hWnd Then
    If Me.Unicode Then
      SendMessageWAsLong Me.hWnd, TTM_ACTIVATE, 0, 0
    Else
      SendMessageAAsLong Me.hWnd, TTM_ACTIVATE, 0, 0
    End If
  End If
End Sub

Public Sub DelTool(ByVal hWndTool As Long)
  Dim DataA As TOOLINFOA
  Dim DataW As TOOLINFOW

  If Me.hWnd Then
    If Me.Unicode Then
      With DataW
        .cbSize = LenB(DataW) - LenB(.lpReserved)
        .hWnd = Me.hWndParent
        .uId = hWndTool
      End With
      SendMessageW Me.hWnd, TTM_DELTOOLW, 0, DataW
    Else
      With DataA
        .cbSize = LenB(DataA) - LenB(.lpReserved)
        .hWnd = Me.hWndParent
        .uId = hWndTool
      End With
      SendMessageA Me.hWnd, TTM_DELTOOLA, 0, DataA
    End If
  End If
End Sub

' destroys the frame
Public Sub Destroy()
  UnSubclassParent
  If prophWnd Then
    UnSubclass

    If Not bAttached Then
      ShowWindow prophWnd, SW_HIDE
      SetParent prophWnd, 0

      DestroyWindow prophWnd
      prophWnd = 0
    End If
  End If
End Sub

' detaches this object from the tooltip control (must have been previously attached)
Public Sub Detach()
  If bAttached Then
    UnSubclass
    UnSubclassParent
    prophWnd = 0

    bAttached = False
  End If
End Sub

#If FullFeatured Or Debuging Then
  Public Sub Hide(Optional ByVal hWndTool As Long = 0)
    Dim DataA As TOOLINFOA
    Dim DataW As TOOLINFOW

    If Me.hWnd Then
      If Me.Unicode Then
        If hWndTool Then
          With DataW
            .cbSize = LenB(DataW) - LenB(.lpReserved)
            .hWnd = Me.hWndParent
            .uId = hWndTool
            .uFlags = TTF_IDISHWND
          End With
          SendMessageW Me.hWnd, TTM_TRACKACTIVATE, 0, DataW
        Else
          SendMessageWAsLong Me.hWnd, TTM_POP, 0, 0
        End If
      Else
        If hWndTool Then
          With DataA
            .cbSize = LenB(DataA) - LenB(.lpReserved)
            .hWnd = Me.hWndParent
            .uId = hWndTool
            .uFlags = TTF_IDISHWND
          End With
          SendMessageA Me.hWnd, TTM_TRACKACTIVATE, 0, DataA
        Else
          SendMessageAAsLong Me.hWnd, TTM_POP, 0, 0
        End If
      End If
    End If
  End Sub

  Public Sub Show(ByVal hWndTool As Long)
    Dim DataA As TOOLINFOA
    Dim DataW As TOOLINFOW

    If Me.hWnd Then
      If Me.Unicode Then
        With DataW
          .cbSize = LenB(DataW) - LenB(.lpReserved)
          .hWnd = Me.hWndParent
          .uId = hWndTool
          .uFlags = TTF_IDISHWND
        End With
        SendMessageW Me.hWnd, TTM_TRACKACTIVATE, 1, DataW
      Else
        With DataA
          .cbSize = LenB(DataA) - LenB(.lpReserved)
          .hWnd = Me.hWndParent
          .uId = hWndTool
          .uFlags = TTF_IDISHWND
        End With
        SendMessageA Me.hWnd, TTM_TRACKACTIVATE, 1, DataA
      End If
    End If
  End Sub

  Public Sub TrackTo(ByVal x As Long, ByVal y As Long)
    If Me.hWnd Then
      If Me.Unicode Then
        PostMessageW Me.hWnd, TTM_TRACKPOSITION, 0, MAKELONG(x, y)
      Else
        PostMessageA Me.hWnd, TTM_TRACKPOSITION, 0, MAKELONG(x, y)
      End If
    End If
  End Sub
#End If


#If CustomDraw Or Debuging Then
  Private Sub drawTransparentBitmap(ByVal hBitmap As Long, ByVal MaskClr As Long, ByVal hDC_Dest As Long, ByVal x As Long, ByVal y As Long, ByVal cx As Long, ByVal cy As Long)
    Const SRCAND = &H8800C6
    Const SRCCOPY = &HCC0020
    Const SRCINVERT = &H660046
    Dim hBMP As Long
    Dim hBMP_Mask As Long
    Dim hBMP_Mask_Old As Long
    Dim hBMP_Old As Long
    Dim hBMP_Pic_Old As Long
    Dim hBMP_PicCopy As Long
    Dim hBMP_PicCopy_Old As Long
    Dim hDC_Mem As Long
    Dim hDC_Mem_Mask As Long
    Dim hDC_Mem_Pic As Long
    Dim hDC_Mem_PicCopy As Long

    hDC_Mem = CreateCompatibleDC(hDC_Dest)     ' our working DC
    If hDC_Mem Then
      hBMP = CreateCompatibleBitmap(hDC_Dest, cx, cy)
      If hBMP Then
        hBMP_Old = SelectObject(hDC_Mem, hBMP)
        hDC_Mem_Mask = CreateCompatibleDC(hDC_Dest)     ' our working DC for the mask
        If hDC_Mem_Mask Then
          hBMP_Mask = CreateBitmapAsLong(cx, cy, 1, 1, 0)
          If hBMP_Mask Then
            hBMP_Mask_Old = SelectObject(hDC_Mem_Mask, hBMP_Mask)
            hDC_Mem_Pic = CreateCompatibleDC(hDC_Dest)     ' the DC that holds <hBitmap>
            If hDC_Mem_Pic Then
              hBMP_Pic_Old = SelectObject(hDC_Mem_Pic, hBitmap)
              hDC_Mem_PicCopy = CreateCompatibleDC(hDC_Dest)     ' a DC that holds a "fresh" copy of <hBitmap>
              If hDC_Mem_PicCopy Then
                hBMP_PicCopy = CreateCompatibleBitmap(hDC_Dest, cx, cy)
                If hBMP_PicCopy Then
                  hBMP_PicCopy_Old = SelectObject(hDC_Mem_PicCopy, hBMP_PicCopy)

                  ' copy the target's background into our working DC
                  BitBlt hDC_Mem, 0, 0, cx, cy, hDC_Dest, x, y, SRCCOPY

                  ' create a copy of the bitmap
                  BitBlt hDC_Mem_PicCopy, 0, 0, cx, cy, hDC_Mem_Pic, 0, 0, SRCCOPY

                  ' combine the bitmap several times to get a mask
                  SetBkColor hDC_Mem_PicCopy, MaskClr
                  BitBlt hDC_Mem_Mask, 0, 0, cx, cy, hDC_Mem_PicCopy, 0, 0, SRCCOPY
                  BitBlt hDC_Mem_PicCopy, 0, 0, cx, cy, hDC_Mem_Mask, 0, 0, SRCINVERT
                  BitBlt hDC_Mem, 0, 0, cx, cy, hDC_Mem_Mask, 0, 0, SRCAND
                  BitBlt hDC_Mem, 0, 0, cx, cy, hDC_Mem_PicCopy, 0, 0, SRCINVERT

                  ' transfer the result to the source DC
                  BitBlt hDC_Dest, x, y, cx, cy, hDC_Mem, 0, 0, SRCCOPY

                  SelectObject hDC_Mem_PicCopy, hBMP_PicCopy_Old
                  DeleteObject hBMP_PicCopy
                End If
                DeleteDC hDC_Mem_PicCopy
              End If
              SelectObject hDC_Mem_Pic, hBMP_Pic_Old
              DeleteDC hDC_Mem_Pic
            End If
            SelectObject hDC_Mem_Mask, hBMP_Mask_Old
            DeleteObject hBMP_Mask
          End If
          DeleteDC hDC_Mem_Mask
        End If
        SelectObject hDC_Mem, hBMP_Old
        DeleteObject hBMP
      End If
      DeleteDC hDC_Mem
    End If
  End Sub

  Private Function handleWM_PAINT(ByVal oldWinMain As Long, ByVal hWnd As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    Const PRF_CHILDREN = &H10&
    Const PRF_CLIENT = &H4&
    Const PRF_OWNED = &H20&
    Const SRCCOPY = &HCC0020
    Const WM_PRINTCLIENT = &H318
    Dim Data As PAINTSTRUCT
    Dim hBMP_Content As Long
    Dim hBMP_Content_Old As Long
    Dim hBMP_Old As Long
    Dim hBMP_Text As Long
    Dim hBMP_Text_Old As Long
    Dim hBrBk As Long
    Dim hDC_Mem_BMP As Long
    Dim hDC_Mem_Content As Long
    Dim hDC_Mem_Text As Long
    Dim hDC_TT As Long
    Dim rcRedraw As RECT
    Dim rcText As RECT
    Dim x As Long
    Dim y As Long
    Static bRecursive As Boolean

    If bRecursive Then Exit Function

    bRecursive = True
    BeginPaint hWnd, Data

    hDC_TT = Data.hDC
    rcRedraw = Data.rcPaint

    hDC_Mem_Content = CreateCompatibleDC(hDC_TT)
    If hDC_Mem_Content Then
      With rcContent
        hBMP_Content = CreateCompatibleBitmap(hDC_TT, .Right - .Left, .Bottom - .Top)
      End With
      If hBMP_Content Then
        hBMP_Content_Old = SelectObject(hDC_Mem_Content, hBMP_Content)

        #If FullFeatured Or Debuging Then
          hBrBk = CreateSolidBrush(Me.BackColor)
        #Else
          If Me.Unicode Then
            hBrBk = CreateSolidBrush(SendMessageWAsLong(Me.hWnd, TTM_GETTIPBKCOLOR, 0, 0))
          Else
            hBrBk = CreateSolidBrush(SendMessageAAsLong(Me.hWnd, TTM_GETTIPBKCOLOR, 0, 0))
          End If
        #End If
        If hBrBk Then FillRect hDC_Mem_Content, rcContent, hBrBk

        If Me.UseMaskColor Then
          hDC_Mem_BMP = -1
        Else
          hDC_Mem_BMP = CreateCompatibleDC(hDC_TT)
        End If
        If hDC_Mem_BMP Then
          If hDC_Mem_BMP <> -1 Then hBMP_Old = SelectObject(hDC_Mem_BMP, Me.hBitmap)

          hDC_Mem_Text = CreateCompatibleDC(hDC_TT)
          If hDC_Mem_Text Then
            hBMP_Text = CreateCompatibleBitmap(hDC_TT, TextSize.cx, TextSize.cy)
            If hBMP_Text Then
              hBMP_Text_Old = SelectObject(hDC_Mem_Text, hBMP_Text)

              rcText.Left = 0
              rcText.Right = rcText.Left + TextSize.cx
              rcText.Top = 0
              rcText.Bottom = rcText.Top + TextSize.cy
              If hBrBk Then FillRect hDC_Mem_Text, rcText, hBrBk

              ' let Windows draw the actual tooltip into our memory DC
              If oldWinMain Then
                If Me.Unicode Then
                  CallWindowProcW oldWinMain, hWnd, WM_PRINTCLIENT, hDC_Mem_Text, PRF_CHILDREN Or PRF_CLIENT Or PRF_OWNED
                Else
                  CallWindowProcA oldWinMain, hWnd, WM_PRINTCLIENT, hDC_Mem_Text, PRF_CHILDREN Or PRF_CLIENT Or PRF_OWNED
                End If
              End If

              ' now put it all together into the memory DC
              ' start with the bitmap
              Select Case Me.BitmapPosition
                Case BitmapPositionConstants.bpLeft
                  x = Me.BitmapMarginX
                  y = ((rcContent.Bottom - rcContent.Top) - BMPSize.cy) / 2
                Case BitmapPositionConstants.bpTop
                  x = ((rcContent.Right - rcContent.Left) - BMPSize.cx) / 2
                  y = Me.BitmapMarginY
                Case BitmapPositionConstants.bpRight
                  x = TextSize.cx
                  #If FullFeatured Or Debuging Then
                    If Me.TitleText = "" Then x = x + Me.BitmapMarginX
                  #Else
                    x = x + Me.BitmapMarginX
                  #End If
                  y = ((rcContent.Bottom - rcContent.Top) - BMPSize.cy) / 2
                Case BitmapPositionConstants.bpBottom
                  x = ((rcContent.Right - rcContent.Left) - BMPSize.cx) / 2
                  y = TextSize.cy
                  #If FullFeatured Or Debuging Then
                    If Me.TitleText = "" Then y = y + Me.BitmapMarginY
                  #Else
                    y = y + Me.BitmapMarginY
                  #End If
              End Select
              If Me.UseMaskColor Then
                drawTransparentBitmap Me.hBitmap, Me.MaskColor, hDC_Mem_Content, x, y, BMPSize.cx, BMPSize.cy
              Else
                BitBlt hDC_Mem_Content, x, y, BMPSize.cx, BMPSize.cy, hDC_Mem_BMP, 0, 0, SRCCOPY
              End If

              ' continue with the actual tooltip
              Select Case Me.BitmapPosition
                Case BitmapPositionConstants.bpLeft
                  x = Me.BitmapMarginX + BMPSize.cx
                  #If FullFeatured Or Debuging Then
                    If Me.TitleText = "" Then x = x + Me.BitmapMarginX
                  #Else
                    x = x + Me.BitmapMarginX
                  #End If
                  y = ((rcContent.Bottom - rcContent.Top) - TextSize.cy) / 2
                Case BitmapPositionConstants.bpTop
                  x = ((rcContent.Right - rcContent.Left) - TextSize.cx) / 2
                  y = Me.BitmapMarginY + BMPSize.cy
                  #If FullFeatured Or Debuging Then
                    If Me.TitleText = "" Then y = y + Me.BitmapMarginY
                  #Else
                    y = y + Me.BitmapMarginY
                  #End If
                Case BitmapPositionConstants.bpRight
                  x = 0
                  #If FullFeatured Or Debuging Then
                    If Me.TitleText = "" Then x = 3
                  #Else
                    x = 3
                  #End If
                  y = ((rcContent.Bottom - rcContent.Top) - TextSize.cy) / 2
                Case BitmapPositionConstants.bpBottom
                  x = ((rcContent.Right - rcContent.Left) - TextSize.cx) / 2
                  y = 0
                  #If FullFeatured Or Debuging Then
                    If Me.TitleText = "" Then y = 3
                  #Else
                    y = 3
                  #End If
              End Select
              BitBlt hDC_Mem_Content, x, y, TextSize.cx, TextSize.cy, hDC_Mem_Text, 0, 0, SRCCOPY

              ' now copy the desired rectangle into the target DC
              With rcRedraw
                BitBlt hDC_TT, .Left, .Top, .Right - .Left, .Bottom - .Top, hDC_Mem_Content, .Left, .Top, SRCCOPY
              End With
              handleWM_PAINT = 0

              SelectObject hDC_Mem_Text, hBMP_Text_Old
              DeleteObject hBMP_Text
            End If
            DeleteDC hDC_Mem_Text
          End If

          If hDC_Mem_BMP <> -1 Then
            SelectObject hDC_Mem_BMP, hBMP_Old
            DeleteDC hDC_Mem_BMP
          End If
        End If

        If hBrBk Then DeleteObject hBrBk
        SelectObject hDC_Mem_Content, hBMP_Content_Old
        DeleteObject hBMP_Content_Old
      End If
      DeleteDC hDC_Mem_Content
    End If

    EndPaint hWnd, Data
    bRecursive = False
  End Function
#End If

#If FullFeatured Or Debuging Then
  ' concatenates <Lo> and <Hi> to a 32-bit-number
  Private Function MAKELONG(ByVal Lo As Integer, ByVal Hi As Integer) As Long
    Dim ret As Long

    CopyMemory ByVal VarPtr(ret), ByVal VarPtr(Lo), LenB(Lo)
    CopyMemory ByVal VarPtr(ret) + LenB(Lo), ByVal VarPtr(Hi), LenB(Hi)

    MAKELONG = ret
  End Function
#End If

' cuts <txt> at the first occurance of Chr(0)
Private Function removeNull(ByVal txt As String) As String
  Dim pos As Integer

  pos = InStr(txt, Chr(0))
  If pos Then txt = Left(txt, pos - 1)
  removeNull = txt
End Function

#If CustomDraw Or Debuging Then
  Private Sub resizeToolTipWithBitmap()
    Const SWP_NOACTIVATE = &H10
    Const SWP_NOMOVE = &H2
    Const SWP_NOOWNERZORDER = &H200
    Const SWP_NOREPOSITION = SWP_NOOWNERZORDER
    Const SWP_NOZORDER = &H4
    Dim rcClient As RECT
    Dim szNeeded As SIZE

    ' get size without bitmap
    GetClientRect Me.hWnd, rcClient
    TextSize.cx = rcClient.Right - rcClient.Left
    TextSize.cy = rcClient.Bottom - rcClient.Top

    ' calculate the desired size
    Select Case Me.BitmapPosition
      Case BitmapPositionConstants.bpLeft, BitmapPositionConstants.bpRight
        szNeeded.cx = Me.BitmapMarginX + BMPSize.cx + TextSize.cx
        #If FullFeatured Or Debuging Then
          If Me.TitleText = "" Then szNeeded.cx = szNeeded.cx + Me.BitmapMarginX
        #Else
          szNeeded.cx = szNeeded.cx + Me.BitmapMarginX
        #End If
        If (BMPSize.cy + 2 * Me.BitmapMarginY) > TextSize.cy Then
          szNeeded.cy = BMPSize.cy + 2 * Me.BitmapMarginY
        Else
          szNeeded.cy = TextSize.cy
        End If

      Case BitmapPositionConstants.bpTop, BitmapPositionConstants.bpBottom
        szNeeded.cy = Me.BitmapMarginY + BMPSize.cy + TextSize.cy
        #If FullFeatured Or Debuging Then
          If Me.TitleText = "" Then szNeeded.cy = szNeeded.cy + Me.BitmapMarginY
        #Else
          szNeeded.cy = szNeeded.cy + Me.BitmapMarginY
        #End If
        If (BMPSize.cx + 2 * Me.BitmapMarginX) > TextSize.cx Then
          szNeeded.cx = BMPSize.cx + 2 * Me.BitmapMarginX
        Else
          szNeeded.cx = TextSize.cx
        End If
    End Select

    ' don't forget we've a border of 1 pixel
    ' TODO: get border size from system settings
    szNeeded.cx = szNeeded.cx + 2
    szNeeded.cy = szNeeded.cy + 2

    ' apply the size
    SetWindowPos Me.hWnd, 0, 0, 0, szNeeded.cx, szNeeded.cy, SWP_NOACTIVATE Or SWP_NOMOVE Or SWP_NOREPOSITION Or SWP_NOZORDER
    GetClientRect Me.hWnd, rcContent
  End Sub
#End If

Private Sub Subclass()
  #If CustomDraw Or FireEvents Or Debuging Then
    If Me.Unicode Then
      oldWinMain = SetWindowLongPtrW(Me.hWnd, GWLP_WNDPROC, VarPtr(mASM(0)))
    Else
      oldWinMain = SetWindowLongPtrA(Me.hWnd, GWLP_WNDPROC, VarPtr(mASM(0)))
    End If
  #End If
End Sub

Private Sub SubclassParent()
  #If CustomDraw Or FireEvents Or Debuging Then
    If Me.Unicode Then
      oldWinMainParent = SetWindowLongPtrW(Me.hWndParent, GWLP_WNDPROC, VarPtr(mASM(0)))
    Else
      oldWinMainParent = SetWindowLongPtrA(Me.hWndParent, GWLP_WNDPROC, VarPtr(mASM(0)))
    End If
  #End If
End Sub

#If FullFeatured Or Debuging Then
  ' transforms an OLE color to a RGB color
  Private Function TranslateColor(ByVal Clr As OLE_COLOR, Optional ByVal hPal As Long = 0) As Long
    If OleTranslateColor(Clr, hPal, TranslateColor) Then TranslateColor = CLR_INVALID
  End Function
#End If

Private Sub UnSubclass()
  If oldWinMain Then
    If Me.Unicode Then
      SetWindowLongPtrW Me.hWnd, GWLP_WNDPROC, oldWinMain
    Else
      SetWindowLongPtrA Me.hWnd, GWLP_WNDPROC, oldWinMain
    End If
    oldWinMain = 0
  End If
End Sub

Private Sub UnSubclassParent()
  If oldWinMainParent Then
    If Me.Unicode Then
      SetWindowLongPtrW Me.hWndParent, GWLP_WNDPROC, oldWinMainParent
    Else
      SetWindowLongPtrA Me.hWndParent, GWLP_WNDPROC, oldWinMainParent
    End If
    oldWinMainParent = 0
  End If
End Sub

#If CustomDraw Or FireEvents Or Debuging Then
  Private Function WinMain_Parent(ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    #If CustomDraw Or FullFeatured Or Debuging Then
      Const WM_NOTIFY = &H4E
    #End If
    Dim callOldWinMain As Boolean
    #If FullFeatured Or Debuging Then
      Dim bStore As Boolean
      Dim DispDataA As NMTTDISPINFOA
      Dim DispDataW As NMTTDISPINFOW
    #End If
    #If CustomDraw Or FullFeatured Or Debuging Then
      Dim NotData As NMHDR
    #End If
    Dim ret As Long
    #If FullFeatured Or Debuging Then
      Dim str As String
    #End If

    callOldWinMain = True
    #If CustomDraw Or FullFeatured Or Debuging Then
      Select Case Msg
        Case WM_NOTIFY
          CopyMemory NotData, ByVal lParam, LenB(NotData)
          If NotData.hwndFrom = Me.hWnd Then
            Select Case NotData.code
              #If FullFeatured Or Debuging Then
                Case TTN_LINKCLICK
                  RaiseEvent ClickedLink
                Case TTN_NEEDTEXTA
                  If Me.Unicode Then
                    RaiseEvent NeedText(str, bStore)
                    If str <> "" Then
                      CopyMemory DispDataW, ByVal lParam, LenB(DispDataW)
                      With DispDataW
                        If Len(str) <= 80 Then
                          lstrcpynWAsLong VarPtr(.szText(0)), str, 80
                        Else
                          .lpszText = StrPtr(str)
                        End If
                        .hinst = 0
                        If bStore Then .uFlags = .uFlags Or TTF_DI_SETITEM
                      End With
                      CopyMemory ByVal lParam, DispDataW, LenB(DispDataW)
                    End If
                  Else
                    RaiseEvent NeedText(str, bStore)
                    If str <> "" Then
                      CopyMemory DispDataA, ByVal lParam, LenB(DispDataA)
                      With DispDataA
                        If Len(str) <= 80 Then
                          lstrcpynAAsLong VarPtr(.szText(0)), str, 80
                        Else
                          .lpszText = StrPtr(str)
                        End If
                        .hinst = 0
                        If bStore Then .uFlags = .uFlags Or TTF_DI_SETITEM
                      End With
                      CopyMemory ByVal lParam, DispDataA, LenB(DispDataA)
                    End If
                  End If
                Case TTN_NEEDTEXTW
                  RaiseEvent NeedText(str, bStore)
                  If str <> "" Then
                    CopyMemory DispDataW, ByVal lParam, LenB(DispDataW)
                    With DispDataW
                      If Len(str) <= 80 Then
                        lstrcpynWAsLong VarPtr(.szText(0)), str, 80
                      Else
                        .lpszText = StrPtr(str)
                      End If
                      .hinst = 0
                      If bStore Then .uFlags = .uFlags Or TTF_DI_SETITEM
                    End With
                    CopyMemory ByVal lParam, DispDataW, LenB(DispDataW)
                  End If
                Case TTN_POP
                  RaiseEvent BeforeHide
              #End If
              Case TTN_SHOW
                #If CustomDraw Or Debuging Then
                  If Me.hBitmap Then resizeToolTipWithBitmap
                #End If

                #If FullFeatured Or Debuging Then
                  RaiseEvent BeforeShow
                #End If
            End Select
          End If
      End Select
    #End If

    If callOldWinMain Then
      If Me.Unicode Then
        ret = CallWindowProcW(oldWinMainParent, hWnd, Msg, wParam, lParam)
      Else
        ret = CallWindowProcA(oldWinMainParent, hWnd, Msg, wParam, lParam)
      End If
    End If

    WinMain_Parent = ret
  End Function

  Private Function WinMain_ToolTip(ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    #If CustomDraw Or Debuging Then
      Const WM_PAINT = &HF
    #End If
    Dim callOldWinMain As Boolean
    Dim ret As Long

    callOldWinMain = True
    #If FullFeatured Or Debuging Then
      RaiseEvent ProcessMessage(oldWinMain, hWnd, Msg, wParam, lParam, callOldWinMain, ret)
    #End If

    If callOldWinMain Then
      #If CustomDraw Or Debuging Then
        Select Case Msg
          Case WM_PAINT
            If Me.hBitmap Then
              WinMain_ToolTip = handleWM_PAINT(oldWinMain, hWnd, wParam, lParam)
              Exit Function
            End If
        End Select
      #End If

      If Me.Unicode Then
        ret = CallWindowProcW(oldWinMain, hWnd, Msg, wParam, lParam)
      Else
        ret = CallWindowProcA(oldWinMain, hWnd, Msg, wParam, lParam)
      End If
    End If

    WinMain_ToolTip = ret
  End Function
#End If
